#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/*
 * Core i7 1160G7
 *  L1d 48KiB 12w64s = 768 cache lines,   max N = 768 * 8 = 6144
 *  L2 1280KiB 20w1024s = 20480 lines,    max N = 163840
 *  L3 12288KiB 12w16384s = 196608 lines, max N = 1572864
 */

#ifndef N
#define N 700000
#endif

#define NREPS 1000

double a[N];
double b[N];
double c[N];

double wtime()
{
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return ts.tv_sec + ts.tv_nsec * 1E-9;
}

void vec_init(double a[N], double b[N], double c[N])
{
    srand(0);
    for (int i = 0; i < N; i++) {
        a[i] = rand() % 100;
        b[i] = rand() % 100;
        c[i] = rand() % 100;
    }
}

void vec_sum(double a[N], double b[N], double c[N])
{
    for (int i = 0; i < N; i++) {
        b[i] = a[i] + 1.0;
    }

    /* For large N, the initial elems of the a[i], b[i] will be evicted */
    for (int i = 0; i < N; i++) {
        c[i] = b[i] * 4.0;
    }
}

void vec_sum_fusion(double a[N], double b[N], double c[N])
{
    for (int i = 0; i < N; i++) {
        b[i] = a[i] + 1.0;
        c[i] = b[i] * 4.0; /* uses cached b[i] */
    }
}


int main()
{
    double t1, t2;

    #if 1
    vec_init(a, b, c);
    t1 = wtime();
    for (int i = 0; i < NREPS; i++)
        vec_sum(a, b, c);
    t1 = wtime() - t1;
    t1 /= NREPS;
    printf("# Vec sum def: N=%d, elapsed time (sec) %.6f, Melems/sec %.1f\n", N, t1, N * 1E-6 / t1);
    #endif

    #if 0
    vec_init(a, b, c);
    t2 = wtime();
    for (int i = 0; i < NREPS; i++)
        vec_sum_fusion(a, b, c);
    t2 = wtime() - t2;
    t2 /= NREPS;
    printf("# Vec sum fusion: N=%d, elapsed time (sec) %.6f, Melems/sec %.1f\n", N, t2, N * 1E-6 / t2);
    #endif

    return 0;
}
